from django.db.models import Q
from functools import reduce


class Query:

    def __init__(self, conditions: list, orders=None):
        self.is_empty = conditions is None or len(conditions) == 0
        self.orders = orders or {'sort': -1}
        self.conditions = [{'includes': {}, 'excludes': {}}] if self.is_empty else conditions

    def orm(self):
        orders = __translate_orders__(self.orders)
        includes = __translate_orm_condition__(self.conditions, 'includes')
        excludes = __translate_orm_condition__(self.conditions, 'excludes')
        return includes, excludes, orders

    def mon(self):
        finder = {}
        return finder, self.orders


def __fun__(x, y): return x | y


def __translate_orders__(orders):
    array = []
    for key in orders:
        value = orders.get(key)
        prefix = '-' if value == -1 else ''
        array.append(f'{prefix}{key}')
    return array


def __translate_orm_condition__(conditions, _type):
    return reduce(__fun__, [Q(**cond.get(_type, {})) for cond in conditions])



from deepmerge import always_merger
from django.db.models import ManyToOneRel, ForeignKey, ManyToManyRel, ManyToManyField, OneToOneField, OneToOneRel
from django.db.models import IntegerField, BooleanField, FloatField, TextField, CharField
from django.db.models import FileField, JSONField
from django.db.models import DateTimeField, TimeField, DateField
from django.db.models.options import Options

from ..dao.abstract import AbstractField
from ..dao.frame import meta_field_domain_mapping, meta_field_tool_mapping
from ..models.core import VTree


class OrmField(AbstractField):

    def __init__(self, entity, model_field, is_tree):
        self.db = 'orm'
        self.entity = entity
        self.model_field = model_field
        self.is_tree = is_tree
        self.not_null = not model_field.null
        self.clazz = type(model_field)
        self.multiple = self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel]
        self.prop = self.__prop__()
        self.domain = self.__domain__()
        self.model = self.__model__()
        self.label = self.__label__()
        self.tool = meta_field_domain_mapping.get(self.domain, 'none')
        self.column_width = column_width(self.domain, self.tool)
        self.refer = self.__refer__()
        self.format = self.__formating__()

    def to_dict(self):
        _field = {
            "prop": self.prop,
            "label": self.label,
            "name": self.label,
            "domain": self.domain,
            "refer": self.refer,
            "format": self.format,
            "not_null": self.not_null,
            "column_width": self.column_width,
            "tool": self.tool,
            "disabled": False,
        }
        if self.is_tree:
            if self.prop in ['pid', 'isLeaf']:
                # _field['hide_on_table'] = True
                _field['hide_on_form'] = True
                _field['hide_on_form_branch'] = True
                _field['hide_on_form_leaf'] = True
            elif self.prop in ['icon']:
                _field['tool'] = 'icon'
        elif self.refer['isTree']:
            _field['tool'] = 'tree'
        return _field

    def __label__(self):
        return self.model._meta.verbose_name \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel, OneToOneRel, OneToOneField] \
            else self.model_field.verbose_name

    def __model__(self):
        return self.model_field.related_model \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel, OneToOneRel, OneToOneField, ForeignKey] \
            else None

    def __domain__(self):
        return self.clazz.__name__ \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel, OneToOneRel, OneToOneField] \
            else self.model_field.get_internal_type()

    def __prop__(self):
        return self.model_field.name + "_id" \
            if self.clazz in [ForeignKey, OneToOneRel, OneToOneField] \
            else self.model_field.name

    def __refer__(self):
        refer = {
            "entity": None, "db": self.db,
            "value": "id", "label": 'name', "display": "id",
            "multiple": self.multiple, "strict": False, "remote": False,
            "includes": {}, "excludes": {},
            "root": 0, "isTree": False
        }
        if self.model:
            meta: Options = getattr(self.model, '_meta')

            refer['entity'] = '%s.%s' % (meta.app_label, self.model.__name__)
            refer['isTree'] = issubclass(self.model, VTree)
        return refer

    def __formating__(self):
        _format = {
            # 文本
            "maxlength": None,
            "type": 'text',

            # 数值
            "min": None,
            "max": None,
            "step": 1,
            "precision": None,
            "step_strictly": False,

            # 日期
            "frequency": "date",

            # 文件
            "maximum": 5,
            "accept": [],
            "width": 800,
            "height": None,
            "file_name_field": None,
            "locked": False,

            # 集合
            "set": {},
            "multiple": False
        }
        if self.clazz == CharField:
            _format['maxlength'] = self.model_field.max_length
            _format['$maxlength'] = self.model_field.max_length
        if self.clazz == TextField:
            _format['type'] = "textarea"
        elif self.clazz == DateTimeField:
            _format['frequency'] = "datetime"
        elif self.clazz == IntegerField:
            _format['precision'] = 0
            _format['step_strictly'] = True
        return _format


def column_width(domain, tool):
    if domain in ['BooleanField', 'FileField', 'JSONField']:
        return 80
    elif domain in ['DateField', 'DateTimeField', 'TimeField']:
        return 120
    elif domain in ['ManyToManyRel', 'ManyToManyRel', 'ManyToOneRel']:
        return 80
    elif domain in ['TextField']:
        return 80 if tool == 'rich' else 0
    return 0



import asyncio

from ..channels.sender import ValarChannelSender


async def execute_channel(method, sender: ValarChannelSender):
    thread = asyncio.to_thread(__execute__, method, sender)
    asyncio.create_task(thread)


def __execute__(method, sender: ValarChannelSender):
    sender.start()
    method(sender)
    sender.stop()



import json

from minio import Minio

from ..classes.valar_response import ValarResponse
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.http import QueryDict

from ..dao import Dao, OrmDao
from ..dao.engine import ValarEngine


def save_file(request, db, entity):
    params: QueryDict = request.POST.dict()
    _id, prop, field = (params.get(key) for key in ['id', 'prop', 'field'])
    file: InMemoryUploadedFile = request.FILES['file']
    dao = Dao(entity, db)
    engine = ValarEngine().get_minio_bucket(entity)
    item = dao.find_one(_id)
    if engine is None or item is None:
        return ValarResponse(False)
    old_value = getattr(item, prop)
    if old_value:
        engine.remove(old_value.name)
    object_name = engine.get_object_name(_id, prop, file.name)
    path = engine.upload(object_name, file.read())
    setattr(item, prop, path)
    if field:
        setattr(item, field, file.name)
    item.save()
    return ValarResponse(path)


def remove_file(request, db, entity):
    body = json.loads(request.body)
    _id, prop, field = (body.get(key) for key in ['id', 'prop', 'field'])
    dao = Dao(entity, db)
    item = dao.find_one(_id)
    engine = ValarEngine().get_minio_bucket(entity)
    if engine is None or item is None:
        return ValarResponse(False)
    old_value = getattr(item, prop)
    if old_value:
        engine.remove(old_value.name)
    setattr(item, prop, None)
    if field:
        setattr(item, field, None)
    item.save()
    return ValarResponse(True)



import datetime
import json

from django.core.paginator import Paginator

from .abstract import AbstractDao
from .query import Query
from django.db.models import ManyToOneRel, ForeignKey, ManyToManyRel, ManyToManyField, OneToOneField, OneToOneRel
from django.db.models import QuerySet
from django.db.models import Manager
from django.db.models.fields.files import FieldFile
from django.forms import FileField

from ..dao.orm_field import OrmField
from ..models.core import VTree, VModel
from ..models.meta import MetaField


class OrmDao(AbstractDao):

    def __init__(self, entity):
        model = self.engine.get_orm_model(entity)
        meta = getattr(model, '_meta')
        self.minio = self.engine.get_minio_bucket(entity)
        self.db = 'orm'
        self.entity = entity
        self.name = meta.verbose_name
        self.is_tree = issubclass(model, VTree)
        self.manager: Manager = model.objects
        fields = {}
        model_fields = meta.get_fields()
        for model_field in model_fields:
            meta_field = OrmField(entity, model_field, self.is_tree)
            fields[meta_field.prop] = meta_field
        self.fields = fields

    def save_many(self, array: list):
        pass

    def values(self, conditions, props: list):
        query_set, _ = self.find(conditions)
        if 'id' not in props:
            props.append('id')
        return list(query_set.values(*props))

    def save_one(self, item, with_id=False):
        oid, simple_item, complex_item = self.__detach_item__(item)
        query_set = self.manager.filter(id=oid) if oid else []
        if len(query_set):
            simple_item['modify_time'] = datetime.datetime.now()
            query_set.update(**simple_item)
            bean = query_set.first()
        else:
            if oid:
                simple_item.update({'id': oid})
            bean = self.manager.create(**simple_item)
            if simple_item.get('sort') is None:
                bean.sort = bean.id
                bean.save()
        self.__save_complex_field__(complex_item, bean)
        bean.save()
        return bean

    def delete_one(self, _id):
        oid = self.object_id(_id)
        flag = oid is not None
        if flag:
            query_set = self.manager.filter(id=oid)
            paths = self.__get_file_paths__(query_set)
            for path in paths:
                self.minio.remove(path)
            query_set.delete()
        return flag

    def find_one(self, _id):
        oid = self.object_id(_id)
        return self.manager.filter(id=oid).first() if oid is not None else None

    def find(self, conditions=None, orders=None, size=0, page=1):
        includes, excludes, orders = Query(conditions, orders).orm()
        query_set = self.manager.filter(includes).exclude(excludes).order_by(*orders)
        total = query_set.count()
        if size:
            paginator = Paginator(query_set, size)
            query_set = paginator.page(page).object_list
        return query_set, total

    def update(self, template, conditions):
        flag = template is not None and len(template.keys())
        if flag:
            oid, simple_item, complex_item = self.__detach_item__(template)
            query_set, total = self.find(conditions)
            query_set.update(**simple_item)
        return flag

    def delete(self, conditions=None) -> list:
        query_set, total = self.find(conditions)
        query_set.delete()
        return self.__get_file_paths__(query_set)

    def serialize(self, o, code=None):
        return self.__to_dict__(o, code) if isinstance(o, QuerySet) else o.full()

    def tree(self, root, conditions=None):
        all_set, _ = self.find()
        if Query(conditions).is_empty:
            return all_set
        values = all_set.values('id', 'pid')
        mapping = {item['id']: item['pid'] for item in values}
        results, _ = self.find(conditions)
        id_set = {root}
        for item in results:
            _id = item.id
            route = []
            while _id is not None:
                route.append(_id)
                _id = mapping.get(_id)
            if root in route:
                id_set.update(route)
        return all_set.filter(id__in=id_set).order_by('-sort')

    """ 以下为私有方法 """

    def __detach_item__(self, item):
        _id = item.get('id')
        if _id:
            del item['id']
        simple_item = {}
        complex_item = {}
        for prop in item:
            meta_field = self.fields.get(prop)
            if meta_field:
                value = item.get(prop)
                if meta_field.domain in ['ManyToOneRel', 'ManyToManyField', 'ManyToManyRel']:
                    complex_item[prop] = value
                elif meta_field.domain in ['OneToOneRel', 'OneToOneField', 'FileField']:
                    complex_item[prop] = value
                else:
                    simple_item[prop] = value
        return self.object_id(_id), simple_item, complex_item

    def __save_complex_field__(self, complex_item, bean):
        for prop in complex_item:
            value = complex_item[prop]
            model_field = self.fields[prop].model_field
            clazz = type(model_field)
            if clazz == ManyToManyField:
                m2m = getattr(bean, prop)
                m2m.clear()
                m2m.add(*value)
            elif clazz == ManyToOneRel:
                getattr(bean, model_field.get_accessor_name()).clear()
                remote_model: VModel = model_field.related_model
                new_set: QuerySet = remote_model.objects.filter(id__in=value)
                remote_field: ForeignKey = model_field.remote_field
                k = remote_field.get_attname()
                new_set.update(**{k: bean.id})
            elif clazz == ManyToManyRel:
                getattr(bean, model_field.get_accessor_name()).clear()
                remote_model: VModel = model_field.related_model
                remote_items: QuerySet = remote_model.objects.filter(id__in=value)
                remote_field: ManyToManyField = model_field.remote_field
                remote_field_prop = remote_field.get_attname()
                for _bean in remote_items:
                    bean_set = getattr(_bean, remote_field_prop)
                    bean_set.add(bean)
            elif clazz == OneToOneRel and value is not None:
                remote_model: VModel = model_field.related_model
                remote_field: OneToOneField = model_field.remote_field
                remote_field_prop = remote_field.get_attname()
                _bean = remote_model.objects.get(id=value)
                __bean = remote_model.objects.filter(**{remote_field_prop: bean.id}).first()
                if __bean:
                    setattr(__bean, remote_field_prop, None)
                    __bean.save()
                setattr(_bean, remote_field_prop, bean.id)
                _bean.save()
            elif clazz == OneToOneField and value is not None:
                __bean = model_field.model.objects.filter(**{prop: value}).first()
                if __bean:
                    setattr(__bean, prop, None)
                    __bean.save()
                setattr(bean, prop, value)
            elif clazz == FileField:
                file_name, _bytes = value
                field_file: FieldFile = getattr(bean, prop)
                if field_file:
                    path = field_file.name
                    self.minio.remove(path)
                object_name = self.minio.get_object_name(bean.id, prop, file_name)
                path = self.minio.upload(object_name, _bytes) if _bytes else None
                setattr(bean, prop, path)

    def __get_file_paths__(self, query_set: QuerySet):
        props = self.props('FileField')
        if len(props):
            items = query_set.values(*props)
            paths = []
            for item in items:
                paths += [i for i in item.values() if i]
            return [path for path in paths if path]
        return []

    def __to_dict__(self, query_set: QuerySet, code=None):
        meta_fields = self.fields.values()
        # 简单字段取值
        simple_props = [field.prop for field in meta_fields if field.domain not in __referred_domains__]
        custom_props = __get_custom_props__(self.entity, code)
        results = list(query_set.filter().values(*[*simple_props, *custom_props]))
        __set_simple_values__(meta_fields, results)
        # 关系型字段取值
        mapping = {row['id']: row for row in results}
        referred_fields = [field for field in meta_fields if field.domain in __referred_domains__]
        pks = mapping.keys()
        for meta_field in referred_fields:
            manager: Manager = query_set.model.objects
            qs = manager.filter(id__in=pks)
            __linkage__(meta_field, qs, mapping)
        return results


""" 以下为静态方法和变量 """

__multiple_domains__ = ['ManyToOneRel', 'ManyToManyField', 'ManyToManyRel']
__referred_domains__ = [*__multiple_domains__, 'OneToOneRel', 'OneToOneField', 'ForeignKey']
__omit_field_props__ = ['create_time', 'modify_time', 'saved', 'sort']
__data_props_formatting__ = {'DateField': '%Y-%m-%d', 'DateTimeField': '%Y-%m-%d %H:%M:%S', 'TimeField': '%H:%M:%S'}


def __get_custom_props__(entity, code='default'):
    field_set = MetaField.objects.filter(view__code=code, view__meta__entity=entity, domain='Custom').values('prop')
    return [item['prop'] for item in field_set if item['prop'] is not None]


def __set_simple_values__(fields, values):
    date_props_mapping = {}
    json_props = []
    for field in fields:
        if isinstance(field, OrmField):
            prop = field.prop
            domain = field.domain
        else:
            prop = field.name
            domain = type(field).__name__
        if domain in __data_props_formatting__.keys():
            date_props_mapping[prop] = __data_props_formatting__[domain]
        elif domain == 'JSONField':
            json_props.append(prop)
    for row in values:
        for prop, formating in date_props_mapping.items():
            if row.get(prop):
                row[prop] = row[prop].strftime(formating)
        for prop in json_props:
            row[prop] = json.loads(row[prop]) if type(row[prop]) is str else row[prop]


def __get_related_props__(fields):
    def fun(field): return type(field).__name__ not in __referred_domains__ and field.name not in __omit_field_props__

    return [field.name for field in fields if fun(field)]


def __linkage__(meta_field: OrmField, query_set: QuerySet, mapping):
    model_field = meta_field.model_field
    prop = model_field.name
    multiple = meta_field.domain in __multiple_domains__

    # 获取级联关系的键索引
    ref_prop = f'{prop}__id'
    edges = query_set.exclude(**{f'{ref_prop}__isnull': True}).values('id', ref_prop)
    if multiple:
        related_primary_keys = set()
        results_mapping = {}
        for edge in edges:
            _id, rid = edge['id'], edge[ref_prop]
            related_primary_keys.add(rid)
            array = results_mapping.get(_id, [])
            array.append(rid)
            results_mapping[_id] = array
    else:
        results_mapping = {row['id']: row[ref_prop] for row in edges if row[ref_prop]}
        related_primary_keys = set(results_mapping.values())

    # 获取级联关系从属方的数据
    related_model = model_field.related_model
    related_fields = related_model._meta.get_fields()
    related_props = __get_related_props__(related_fields)
    related_values = list(related_model.objects.filter(id__in=related_primary_keys).values(*related_props))
    __set_simple_values__(related_fields, related_values)
    related_mapping = {item['id']: item for item in related_values}

    # 将从属方的数据绑定在主数据上
    for _id in mapping:
        row = mapping[_id]
        if multiple:
            keys = results_mapping.get(_id, [])
            items = [related_mapping[pid] for pid in keys]
            row[prop] = keys
            row[f'{prop}_set'] = items
        else:
            key = results_mapping.get(_id)
            item = related_mapping.get(key) if key else None
            row[prop] = item
            row[f'{prop}_id'] = key



from django.urls import path

from .channels.views import handel_channel
from .views import rest, meta, file

urlpatterns = [
    path('socket/<str:handler>', handel_channel),
    path('batch', rest.batch),
    path('<str:db>/<str:entity>/save_one', rest.save_one),
    path('<str:db>/<str:entity>/delete_one', rest.delete_one),
    path('<str:db>/<str:entity>/find_one', rest.find_one),
    path('<str:db>/<str:entity>/find', rest.find),
    path('<str:db>/<str:entity>/update', rest.update),
    path('<str:db>/<str:entity>/search', rest.search),
    path('<str:db>/<str:entity>/values', rest.values),
    # path('<str:db>/<str:entity>/tree', rest.tree),

    path('<str:db>/<str:entity>/meta_view', meta.meta_view),
    path('metas', meta.metas),
    path('add_fields', meta.add_fields),
    path('load_customs', meta.load_customs),
    path('get_fields', meta.get_fields),
    path('save_custom', meta.save_custom),
    path('upload_mapping', meta.upload_mapping),

    path('<str:db>/<str:entity>/save_file', file.save_file),
    path('<str:db>/<str:entity>/remove_file', file.remove_file),

]



import json
from io import BytesIO
from django.conf import settings
from urllib3 import BaseHTTPResponse

from ..classes.singleton_meta import SingletonMeta


class ValarMinio(metaclass=SingletonMeta):

    def __init__(self, client, entity):
        self.client = client
        self.bucket_name = f'{settings.BASE_DIR.name}.{entity}'.replace('_', '-').lower()
        if client and not self.client.bucket_exists(self.bucket_name):
            self.client.make_bucket(self.bucket_name)
            self.client.set_bucket_policy(self.bucket_name, self.__generate_policy__())

    @staticmethod
    def get_object_name(_id, prop, file_name):
        return f"{_id}-{prop}-{file_name}"

    def upload(self, object_name, _bytes):
        file_data = BytesIO(_bytes)
        file_size = len(_bytes)  # file.siz
        self.client.put_object(
            bucket_name=self.bucket_name,
            object_name=object_name,
            data=file_data,
            length=file_size
        )
        return f'{self.bucket_name}/{object_name}'

    def remove(self, path):
        if path:
            bucket_name, object_name = path.split('/')
            self.client.remove_object(
                bucket_name=bucket_name,
                object_name=object_name
            )

    def read(self, object_name) -> BytesIO:
        ret: BaseHTTPResponse = self.client.get_object(
            bucket_name=self.bucket_name,
            object_name=object_name
        )
        return BytesIO(ret.read())

    def __generate_policy__(self):
        return json.dumps({
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"AWS": "*"},
                    "Action": "s3:GetBucketLocation",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}"
                },
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"AWS": "*"},
                    "Action": "s3:ListBucket",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}"
                },
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"AWS": "*"},
                    "Action": "s3:GetObject",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}/*"
                },
                {
                    "Sid": "",
                    "Effect": "Allow",
                    "Principal": {"AWS": "*"},
                    "Action": "s3:PutObject",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}/*"
                }
            ]})



from ..channels.counter import Counter
from ..channels.sender import ValarChannelSender
from ..dao import Dao
from ..dao.engine import ValarEngine

import time


def valar_test_handler(sender: ValarChannelSender):
    data = sender.data
    length = data.get('length', 100)
    counter = Counter(length)
    for i in range(length):
        time.sleep(0.1)
        tick = counter.tick()
        tick.update({'name': 'test1'})
        sender.load(tick)


def batch_handler(sender: ValarChannelSender):
    data = sender.data
    entity, db, method = data.get("entity"), data.get("db"), data.get("method")
    dao = Dao(entity, db)
    body = {"method": method}
    if method == 'save_many':
        array = data.get("array", [])
        counter = Counter(array)
        keys = []
        for item in array:
            item.update({'saved': True})
            bean = dao.save_one(item)
            keys.append(bean.id)
            payload = counter.tick()
            sender.load(payload)
        body.update({'keys': keys})
        sender.done(body)
    elif method == 'delete_many':
        conditions = data.get("conditions", [])
        paths = dao.delete(conditions)
        counter = Counter(len(paths))
        minio = ValarEngine().get_minio_bucket(entity)
        for path in paths:
            minio.remove(path)
            payload = counter.tick()
            sender.load(payload)
        sender.done(body)



import json

from .handler import batch_handler
from ..channels.executer import execute_channel
from ..channels.sender import ValarChannelSender
from ..classes.valar_response import ValarResponse
from ..dao import Dao


async def batch(request):
    sender = ValarChannelSender(request)
    await execute_channel(batch_handler, sender)
    return ValarResponse(True)


def save_one(request, db, entity):
    item = json.loads(request.body)
    dao = Dao(entity, db)
    bean = dao.save_one(item)
    item = dao.serialize(bean)
    return ValarResponse(item)


def values(request, db, entity):
    body = json.loads(request.body)
    dao = Dao(entity, db)
    conditions = body.get('conditions', [])
    props = body.get('props', ['id'])
    array = dao.values(conditions, props)
    return ValarResponse(array)


def delete_one(request, db, entity):
    body = json.loads(request.body)
    _id = body['id']
    dao = Dao(entity, db)
    flag = dao.delete_one(_id)
    return ValarResponse(flag)


def find_one(request, db, entity):
    body = json.loads(request.body)
    _id = body['id']
    dao = Dao(entity, db)
    bean = dao.find_one(_id)
    item = dao.serialize(bean)
    return ValarResponse(item)


def find(request, db, entity):
    conditions = json.loads(request.body)
    dao = Dao(entity, db)
    results, _ = dao.find(conditions)
    results = dao.serialize(results)
    return ValarResponse(results)


def update(request, db, entity):
    body = json.loads(request.body)
    conditions = body.get('conditions', [])
    template = body.get('template')
    dao = Dao(entity, db)
    flag = dao.update(template, conditions)
    return ValarResponse(flag)


def search(request, db, entity):
    body = json.loads(request.body)
    _type = body.get('type')
    orders = body.get('orders')
    size = body.get('size')
    page = body.get('page')
    root = body.get('root')
    code = body.get('code', 'default')
    conditions = body.get('conditions')
    dao = Dao(entity, db)

    if _type == 'tree':
        query_set = dao.tree(root, conditions)
        total = query_set.count()
    else:
        query_set, total = dao.find(conditions, orders, size, page)
    results = dao.serialize(query_set, code)

    return ValarResponse({
        "results": results,
        "type": _type,
        "total": total,
        "root": root
    })
